# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: imagebuilder/service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.imagebuilder import payload_pb2 as imagebuilder_dot_payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1aimagebuilder/service.proto\x12\x15\x63loudidl.imagebuilder\x1a\x1aimagebuilder/payload.proto2p\n\x0cImageService\x12`\n\x08GetImage\x12&.cloudidl.imagebuilder.GetImageRequest\x1a\'.cloudidl.imagebuilder.GetImageResponse\"\x03\x90\x02\x01\x42\xd1\x01\n\x19\x63om.cloudidl.imagebuilderB\x0cServiceProtoH\x02P\x01Z/github.com/unionai/cloud/gen/pb-go/imagebuilder\xa2\x02\x03\x43IX\xaa\x02\x15\x43loudidl.Imagebuilder\xca\x02\x15\x43loudidl\\Imagebuilder\xe2\x02!Cloudidl\\Imagebuilder\\GPBMetadata\xea\x02\x16\x43loudidl::Imagebuilderb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'imagebuilder.service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.cloudidl.imagebuilderB\014ServiceProtoH\002P\001Z/github.com/unionai/cloud/gen/pb-go/imagebuilder\242\002\003CIX\252\002\025Cloudidl.Imagebuilder\312\002\025Cloudidl\\Imagebuilder\342\002!Cloudidl\\Imagebuilder\\GPBMetadata\352\002\026Cloudidl::Imagebuilder'
  _IMAGESERVICE.methods_by_name['GetImage']._options = None
  _IMAGESERVICE.methods_by_name['GetImage']._serialized_options = b'\220\002\001'
  _globals['_IMAGESERVICE']._serialized_start=81
  _globals['_IMAGESERVICE']._serialized_end=193
# @@protoc_insertion_point(module_scope)
